"""
Zenodo interactions handling
"""
import json
import logging
import os.path

import requests

from .uploading import upload_with_progress_bar

_LOGGER = logging.getLogger(__name__)


def _get_deposit(deposition_id, zenodo_url, token):
    url_to_hit = "https://{}/api/deposit/depositions/{}".format(
        zenodo_url, deposition_id
    )
    headers_bucket_request = {"Accept": "application/json"}

    _LOGGER.debug("Sending request to: %s", url_to_hit)
    response = requests.get(
        url_to_hit, headers=headers_bucket_request, params={"access_token": token},
    )
    response.raise_for_status()

    return response


def _get_new_version(deposition_id, zenodo_url, token):
    _LOGGER.info("Creating new version for deposition id: %s", deposition_id)

    _LOGGER.debug("Retrieving deposit")
    response = _get_deposit(deposition_id, zenodo_url, token)

    latest_version_deposition_id = response.json()["links"]["latest"].split("/")[-1]
    _LOGGER.debug("Latest version of record: %s", latest_version_deposition_id)

    url_to_hit = "https://{}/api/deposit/depositions/{}/actions/newversion".format(
        zenodo_url, latest_version_deposition_id
    )

    _LOGGER.debug("Posting to: %s", url_to_hit)

    response = requests.post(url_to_hit, params={"access_token": token},)
    response.raise_for_status()

    new_version = response.json()["links"]["latest_draft"].split("/")[-1]

    _LOGGER.info("New version: %s", new_version)

    return new_version


def _remove_all_files(deposition_id, zenodo_url, token):
    _LOGGER.info("Removing all files at deposition id: %s", deposition_id)

    url_to_hit = "https://{}/api/deposit/depositions/{}/files".format(
        zenodo_url, deposition_id
    )

    _LOGGER.debug("Sending to: %s", url_to_hit)

    response = requests.get(url_to_hit, params={"access_token": token})
    response.raise_for_status()

    for file_entry in response.json():
        _LOGGER.info("Removing file: %s", file_entry["id"])

        url_to_hit = "https://{}/api/deposit/depositions/{}/files/{}".format(
            zenodo_url, deposition_id, file_entry["id"]
        )
        _LOGGER.debug("Posting to: %s", url_to_hit)

        response_file = requests.delete(url_to_hit, params={"access_token": token},)
        response_file.raise_for_status()

    _LOGGER.info("Finished removing all files")


def _set_upload_metadata(deposition_id, zenodo_url, token, deposit_metadata):
    _LOGGER.info("Setting metadata for deposition id: %s", deposition_id)

    url_to_hit = "https://{}/api/deposit/depositions/{}".format(
        zenodo_url, deposition_id
    )

    _LOGGER.debug("Sending to: %s", url_to_hit)

    response = requests.put(
        url_to_hit,
        params={"access_token": token},
        data=json.dumps(deposit_metadata),
        headers={"Content-Type": "application/json"},
    )
    response.raise_for_status()

    _LOGGER.debug("Successfully set metdata")


def create_new_zenodo_version(deposition_id, zenodo_url, token, deposit_metadata):
    """
    Create a new version of a Zenodo record (i.e. a specific deposition ID)

    Parameters
    ----------
    deposition_id : str
        Deposition ID of any DOI which represents a specific version of the record, but crucially **not** the DOI which represents all versions of the record (this won't work).

    zenodo_url : str
        Zenodo url to upload the file to (e.g. ``sandbox.zenodo.org`` or
        ``zenodo.org``)

    token : str
        Token to use to authenticate the request

    deposit_metadata : str
        Path to file containing metadata for this new version

    Returns
    -------
    str
        The deposition ID of the new version of the record
    """
    with open(deposit_metadata, "r") as fileh:
        deposit_metadata_loaded = json.load(fileh)
    # validate deposit_metadata_loaded

    new_version = _get_new_version(deposition_id, zenodo_url, token)

    _remove_all_files(new_version, zenodo_url, token)
    _set_upload_metadata(new_version, zenodo_url, token, deposit_metadata_loaded)

    return new_version


def upload_file(filepath, bucket, zenodo_url, token):
    """
    Upload file to Zenodo

    Parameters
    ----------
    filepath : str
        Path to file to upload

    bucket : str
        Bucket to upload the file to

    zenodo_url : str
        Zenodo url to upload the file to (e.g. ``sandbox.zenodo.org`` or
        ``zenodo.org``)

    token : str
        Token to use to authenticate the upload
    """
    _LOGGER.info(
        "Uploading file `%s` to bucket `%s` at `%s`", filepath, bucket, zenodo_url
    )

    upload_url_no_token = "https://{}/api/files/{}/{}?access_token=".format(
        zenodo_url, bucket, os.path.basename(filepath),
    )
    _LOGGER.debug("Upload url: %s", upload_url_no_token)
    upload_url = "{}{}".format(upload_url_no_token, token)

    upload_with_progress_bar(filepath, upload_url)


def get_bucket_id(deposition_id, zenodo_url, token):
    """
    Get bucket ID for a given Zenodo deposition ID

    Parameters
    ----------
    deposition_id : str
        Deposition ID to check

    zenodo_url : str
        Zenodo url to upload the file to (e.g. ``sandbox.zenodo.org`` or
        ``zenodo.org``)

    token : str
        Token to use to authenticate the request

    Returns
    -------
    str
        The bucket associated with ``deposition_id``
    """
    _LOGGER.info("Determining bucket for deposition_id: %s", deposition_id)

    _LOGGER.debug("Retrieving deposit")
    response = _get_deposit(deposition_id, zenodo_url, token)

    bucket = response.json()["links"]["bucket"]
    _LOGGER.debug("Full url for bucket: %s", bucket)

    bucket = bucket.split("/")[-1]
    _LOGGER.info("Successfully retrieved bucket: %s", bucket)

    return bucket



"""
Command line interface(s)
"""
import logging

import click

from .zenodo import create_new_zenodo_version, get_bucket_id, upload_file

DEFAULT_LOG_FORMAT = "{process} {asctime} {levelname}:{name}:{message}"
"""str: Default format used for logging output"""


class ColorFormatter(logging.Formatter):
    """
    Colour formatter for log messages

    A handy little tool for making our log messages look slightly prettier
    """

    colors = {
        "DEBUG": dict(fg="blue"),
        "INFO": dict(fg="green"),
        "WARNING": dict(fg="yellow"),
        "ERROR": dict(fg="red"),
        "EXCEPTION": dict(fg="red"),
        "CRITICAL": dict(fg="red"),
    }

    def format(self, record):
        """
        Format a record so it has pretty colours

        Parameters
        ----------
        record : :obj:`logging.LogRecord`
            Record to format

        Returns
        -------
        str
            Formatted message string
        """
        formatted_message = super().format(record)

        if not record.exc_info:
            level = record.levelname.upper()

            if level in self.colors:
                level_colour = click.style("{}".format(level), **self.colors[level])
                formatted_message = formatted_message.replace(level, level_colour)

        return formatted_message


class ClickHandler(logging.Handler):
    """
    Handler which emits using click when going to stdout
    """

    _use_stderr = True

    def emit(self, record):
        """
        Emit a record

        Parameters
        ----------
        record : :obj:`logging.LogRecord`
            Record to emit
        """
        try:
            msg = self.format(record)
            click.echo(msg, err=self._use_stderr)

        except Exception:  # pragma: no cover # pylint: disable=broad-except
            self.handleError(record)


_default_handler = ClickHandler()
_default_handler.formatter = ColorFormatter(DEFAULT_LOG_FORMAT, style="{")


@click.group(name="openscm-zenodo")
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "EXCEPTION", "CRITICAL"]),
)
def cli(log_level):
    """
    OpenSCM-Zenodo's command-line interface
    """
    netcdf_scm_logger = logging.getLogger("openscm_zenodo")
    netcdf_scm_logger.handlers.append(_default_handler)
    netcdf_scm_logger.setLevel(log_level)

    logging.captureWarnings(True)


_zenodo_url = click.option(
    "--zenodo-url",
    default="sandbox.zenodo.org",
    type=click.Choice(["sandbox.zenodo.org", "zenodo.org"]),
    show_default=True,
    help="Zenodo server to which to upload.",
)
_token = click.option("--token", envvar="ZENODO_TOKEN")
_deposition_id = click.argument("deposition_id")


@cli.command(context_settings={"help_option_names": ["-h", "--help"]})
@_deposition_id
@_zenodo_url
@_token
@click.argument(
    "deposit_metadata",
    type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
)
def create_new_version(deposition_id, zenodo_url, token, deposit_metadata):
    r"""
    Create new version of a Zenodo record (i.e. a specific Zenodo deposition ID).

    ``deposition_id`` is the deposition ID of any existing version DOI, but crucially **not** the DOI which represents all versions of the record (this won't work). The printed value is the deposition ID of the DOI which represents the new, specific version of the record.

    The deposit metadata should be a ``.json`` file. It will be read and
    validated before the new version is created.
    """
    new_version_deposition_id = create_new_zenodo_version(
        deposition_id=deposition_id,
        zenodo_url=zenodo_url,
        token=token,
        deposit_metadata=deposit_metadata,
    )

    click.echo(new_version_deposition_id)


@cli.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument(
    "file_to_upload",
    type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
)
@click.argument("bucket")
@_zenodo_url
@_token
def upload(file_to_upload, bucket, zenodo_url, token):
    r"""
    Upload a file to a Zenodo bucket

    ``file_to_upload`` will be uploaded to the Zenodo bucket specified by
    ``bucket``.
    """
    upload_file(
        filepath=file_to_upload, bucket=bucket, zenodo_url=zenodo_url, token=token
    )


@cli.command(context_settings={"help_option_names": ["-h", "--help"]})
@_deposition_id
@_zenodo_url
@_token
def get_bucket(deposition_id, zenodo_url, token):
    r"""
    Get the bucket associated with a given Zenodo deposition ID (``deposition_id``)

    The ``upload`` command can then be used to upload files to this bucket.

    This command is handy when you know your deposition ID but you've
    forgotten the bucket address.
    """
    bucket_id = get_bucket_id(
        deposition_id=deposition_id, zenodo_url=zenodo_url, token=token
    )

    click.echo(bucket_id)



from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

import versioneer

PACKAGE_NAME = "openscm-zenodo"
DESCRIPTION = "Command-line tool for uploading to zenodo"
KEYWORDS = ["zenodo", "command-line"]

AUTHORS = [
    ("Jared Lewis", "jared.lewis@climate-energy-college.org"),
    ("Zeb Nicholls", "zebedee.nicholls@climate-energy-college.org"),
]
EMAIL = "jared.lewis@climate-energy-college.org"
URL = "https://github.com/openscm/openscm-zenodo"
PROJECT_URLS = {
    "Bug Reports": "https://github.com/openscm/openscm-zenodo/issues",
    "Documentation": "https://OpenSCMZenodo.readthedocs.io/en/latest",
    "Source": "https://github.com/openscm/openscm-zenodo",
}
LICENSE = "3-Clause BSD License"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: BSD License",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]

ENTRY_POINTS = {"console_scripts": ["openscm-zenodo = openscm_zenodo.cli:cli"]}

REQUIREMENTS = ["click", "requests", "tqdm"]
REQUIREMENTS_OPTIONAL = []
REQUIREMENTS_TESTS = [
    "codecov",
    "pytest",
    "pytest-console-scripts",
    "pytest-cov",
] + REQUIREMENTS_OPTIONAL
REQUIREMENTS_DOCS = ["sphinx>2.1", "sphinx_click", "sphinx_rtd_theme"]
REQUIREMENTS_DEPLOY = ["twine>=1.11.0", "setuptools>=41.2", "wheel>=0.31.0"]

REQUIREMENTS_DEV = [
    *["bandit", "black==19.10b0", "flake8", "isort>=5", "pydocstyle", "pylint",],
    *REQUIREMENTS_OPTIONAL,
    *REQUIREMENTS_TESTS,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_DEPLOY,
]

REQUIREMENTS_EXTRAS = {
    "optional": REQUIREMENTS_OPTIONAL,
    "docs": REQUIREMENTS_DOCS,
    "tests": REQUIREMENTS_TESTS,
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": REQUIREMENTS_DEV,
}


SOURCE_DIR = "src"

PACKAGES = find_packages(SOURCE_DIR)  # no exclude as only searching in `src`
PACKAGE_DIR = {"": SOURCE_DIR}
PACKAGE_DATA = {}


README = "README.rst"

# Get the long description from the README file
with open(README, "r") as f:
    README_LINES = ["OpenSCM-Zenodo", "==============", ""]
    add_line = False
    for line in f:
        if line.strip() == ".. sec-begin-long-description":
            add_line = True
        elif line.strip() == ".. sec-end-long-description":
            break
        elif add_line:
            README_LINES.append(line.strip())

if len(README_LINES) < 3:
    raise RuntimeError("Insufficient description given")


class OpenSCMZenodo(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": OpenSCMZenodo})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description="\n".join(README_LINES),
    long_description_content_type="text/x-rst",
    author=", ".join([author[0] for author in AUTHORS]),
    author_email=", ".join([author[1] for author in AUTHORS]),
    url=URL,
    project_urls=PROJECT_URLS,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
    entry_points=ENTRY_POINTS,
)



"""
File uploading handling

Thanks https://gist.github.com/tyhoff/b757e6af83c1fd2b7b83057adf02c139
for the progress bar.
"""
import logging
import os
import os.path

import requests
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

_LOGGER = logging.getLogger(__name__)


"""dict: Configuration for upload progress bar"""
TQDM_KWARGS = dict(unit="B", unit_scale=True, unit_divisor=1024,)


def upload_with_progress_bar(filepath, upload_url):
    """
    Upload file with a progress bar

    Parameters
    ----------
    filepath : str
        Path to file to upload

    upload_url : str
        URL to put the file onto
    """
    _LOGGER.info("Uploading %s to %s", filepath, upload_url)

    file_size = os.stat(filepath).st_size

    with open(filepath, "rb") as file_handle:
        with tqdm(total=file_size, **TQDM_KWARGS) as tqdm_bar:
            wrapped_file = CallbackIOWrapper(tqdm_bar.update, file_handle, "read")
            requests.put(upload_url, data=wrapped_file)
